/*
 * UnreachableStates.cpp
 *
 *  Created on: 23. 3. 2014
 *	  Author: Tomas Pecka
 */

#include "UnreachableStates.h"

#include <exception/AlibException.h>
#include <automaton/FSM/ExtendedNFA.h>
#include <automaton/FSM/CompactNFA.h>
#include <automaton/FSM/EpsilonNFA.h>
#include <automaton/FSM/NFA.h>
#include <automaton/FSM/DFA.h>

#include <set>
#include <map>
#include <queue>

namespace automaton {

namespace properties {

std::set<automaton::State> UnreachableStates::unreachableStates(const Automaton& automaton) {
	std::set<automaton::State> out;
	automaton.getData().Accept((void*) &out, UnreachableStates::UNREACHABLE_STATES);
	return out;
}

template<class T>
std::set<automaton::State> UnreachableStates::unreachableStates( const T & fsm ) {
	// 1a
	std::deque<std::set<automaton::State>> Qi;
	Qi.push_back( std::set<automaton::State>( ) );
	Qi.at( 0 ) = {fsm.getInitialState( )};

	int i = 1;

	// 1bc
	while( true ) {
		Qi.push_back( Qi.at( i - 1 ) );

		for( const auto & p : Qi.at( i - 1 ) )
			for( const auto & transition : fsm.getTransitionsFromState( p ) )
				Qi.at( i ).insert( transition.second.begin(), transition.second.end() );

		if( Qi.at( i ) == Qi.at( i - 1 ) )
			break;

		i = i + 1;
	}

	return Qi.at( i );
}

template std::set<automaton::State> UnreachableStates::unreachableStates( const automaton::EpsilonNFA & fsm );
template std::set<automaton::State> UnreachableStates::unreachableStates( const automaton::NFA & fsm );
template std::set<automaton::State> UnreachableStates::unreachableStates( const automaton::CompactNFA & fsm );
template std::set<automaton::State> UnreachableStates::unreachableStates( const automaton::ExtendedNFA & fsm );

template<>
std::set<automaton::State> UnreachableStates::unreachableStates( const automaton::MultiInitialStateNFA & fsm ) {
	// 1a
	std::deque<std::set<automaton::State>> Qi;
	Qi.push_back( std::set<automaton::State>( ) );
	Qi.at( 0 ) = fsm.getInitialStates( );

	int i = 1;

	// 1bc
	while( true ) {
		Qi.push_back( Qi.at( i - 1 ) );

		for( const auto & p : Qi.at( i - 1 ) )
			for( const auto & transition : fsm.getTransitionsFromState( p ) )
				Qi.at( i ).insert( transition.second.begin(), transition.second.end() );

		if( Qi.at( i ) == Qi.at( i - 1 ) )
			break;

		i = i + 1;
	}

	return Qi.at( i );
}

template<>
std::set<automaton::State> UnreachableStates::unreachableStates( const automaton::DFA & fsm ) {
	// 1a
	std::deque<std::set<automaton::State>> Qi;
	Qi.push_back( std::set<automaton::State>( ) );
	Qi.at( 0 ). insert( fsm.getInitialState( ) );

	int i = 1;

	// 1bc
	while( true ) {
		Qi.push_back( Qi.at( i - 1 ) );

		for( const auto & p : Qi.at( i - 1 ) )
			for( const auto & transition : fsm.getTransitionsFromState( p ) )
				Qi.at( i ).insert( transition.second );

		if( Qi.at( i ) == Qi.at( i - 1 ) )
			break;

		i = i + 1;
	}

	return Qi.at( i );
}

void UnreachableStates::Visit(void* data, const EpsilonNFA& automaton) const {
	std::set<automaton::State> & out = *((std::set<automaton::State>*) data);
	out = std::move(this->unreachableStates(automaton));
}

void UnreachableStates::Visit(void* data, const MultiInitialStateNFA& automaton) const {
	std::set<automaton::State> & out = *((std::set<automaton::State>*) data);
	out = std::move(this->unreachableStates(automaton));
}

void UnreachableStates::Visit(void* data, const NFA& automaton) const {
	std::set<automaton::State> & out = *((std::set<automaton::State>*) data);
	out = std::move(this->unreachableStates(automaton));
}

void UnreachableStates::Visit(void* data, const DFA& automaton) const {
	std::set<automaton::State> & out = *((std::set<automaton::State>*) data);
	out = std::move(this->unreachableStates(automaton));
}

void UnreachableStates::Visit(void* data, const ExtendedNFA& automaton) const {
	std::set<automaton::State> & out = *((std::set<automaton::State>*) data);
	out = std::move(this->unreachableStates(automaton));
}

void UnreachableStates::Visit(void* data, const CompactNFA& automaton) const {
	std::set<automaton::State> & out = *((std::set<automaton::State>*) data);
	out = std::move(this->unreachableStates(automaton));
}

void UnreachableStates::Visit(void*, const DPDA&) const {
	throw exception::AlibException("Unsupported automaton type DPDA");
}

void UnreachableStates::Visit(void*, const SinglePopDPDA&) const {
	throw exception::AlibException("Unsupported automaton type SinglePopDPDA");
}

void UnreachableStates::Visit(void*, const InputDrivenDPDA&) const {
	throw exception::AlibException("Unsupported automaton type InputDrivenDPDA");
}

void UnreachableStates::Visit(void*, const InputDrivenNPDA&) const {
	throw exception::AlibException("Unsupported automaton type InputDrivenNPDA");
}

void UnreachableStates::Visit(void*, const VisiblyPushdownDPDA&) const {
	throw exception::AlibException("Unsupported automaton type VisiblyPushdownDPDA");
}

void UnreachableStates::Visit(void*, const VisiblyPushdownNPDA&) const {
	throw exception::AlibException("Unsupported automaton type VisiblyPushdownNPDA");
}

void UnreachableStates::Visit(void*, const RealTimeHeightDeterministicDPDA&) const {
	throw exception::AlibException("Unsupported automaton type RealTimeHeightDeterministicDPDA");
}

void UnreachableStates::Visit(void*, const RealTimeHeightDeterministicNPDA&) const {
	throw exception::AlibException("Unsupported automaton type RealTimeHeightDeterministicNPDA");
}

void UnreachableStates::Visit(void*, const NPDA&) const {
	throw exception::AlibException("Unsupported automaton type NPDA");
}

void UnreachableStates::Visit(void*, const SinglePopNPDA&) const {
	throw exception::AlibException("Unsupported automaton type SinglePopNPDA");
}

void UnreachableStates::Visit(void*, const OneTapeDTM&) const {
	throw exception::AlibException("Unsupported automaton type OneTapeDTM");
}

const UnreachableStates UnreachableStates::UNREACHABLE_STATES;

} /* namespace properties */

} /* namespace automaton */

